/* $Id: lkst_buffer.h,v 1.8 2002/06/28 11:27:55 indoh Exp $ */
/* linux/include/linux/lkst_buffer.h */

/************************************************************
 *
 * COPYRIGHT (C) HITACHI,LTD. 2001-2002 ALL RIGHTS RESERVED.
 * WRITTEN BY HITACHI SYSTEMS DEVELOPMENT LABORATORY,
 *            HITACHI CENTRAL RESEARCH LABORATORY.
 *
 * Created by K.HATASAKI <keisukeh@crl.hitachi.co.jp>
 *
 ************************************************************/

#ifndef _LINUX_LKST_BUFFER_H
#define _LINUX_LKST_BUFFER_H


#include <linux/lkst.h>
#ifdef __KERNEL__
#include <linux/types.h>
#include <linux/time.h>
#else /* #ifndef __KERNEL__ */
#include <pthread.h>
#include <sys/time.h>
#endif	/* __KERNEL__ */

		/* mode for ioctl(LKST_IOC_BUFFER_SETRMOD) command */
#define RAW	0x01
#define STD	0x02
#define SORTED	0x04

		/* size of lkst_event_record. It must be power of 2 */
		/* This value will be verified in initialization of 
		   LKST in DEBUG mode. */
#define LKST_SIZEOF_LKST_EVENT_RECORD	(64)

		/* 
		 *  following 4 lines should be deleted
		 *  if they has been already defined. 
		 */
typedef int posix_log_recid_t;
typedef int posix_log_facility_t;
typedef int posix_log_severity_t;
typedef int posix_log_procid_t;

#ifdef __KERNEL__
typedef unsigned long int pthread_t;
#endif /*__KERNEL__ */


#ifndef PXLOG_BINARY
#define PXLOG_BINARY	1
#endif /* PXLOG_BINARY */

#ifndef LOG_KERN
#define LOG_KERN	0
#endif /* LOG_KERN */

#ifndef LOG_DEBUG
#define LOG_DEBUG	7
#endif /* LOG_DEBUG */

#define LOGFILE_MAGIC   0xbeefface
#define LOGREC_MAGIC    0xfeefface

#define LKST_ARCH_NAME_LEN 65

/* Logfile Header for evlog compatibility. */
typedef struct log_header {
	uint		log_magic;	/* Magic number indicating log file */
	long		log_version;	/* Event logging version */
	posix_log_recid_t last_recId;	/* Last recid in log: used by evlogd */
	off_t		reserved1;
	uint		log_generation;	/* Changes during log maintenance */
} log_header_t;

struct lkst_log_buffer {
	size_t read_size;		/* size of event records to read */
	lkst_buffer_id id;		/* event buffer ID */
	size_t result_read_size;	/* size of read event records */
	struct timeval xtime;		/* xtime */
	lkst_tsc_t tsc;			/* machine cycle */ 
					/* These two will be used to calculate 
					 * time of events in real time. */
	lkst_cpu_freq_t	cpu_freq;	/* cpu clockspeed in kHz */ 
	struct lkst_log_record *buffer;
			/* address of a buffer to store event records */

	int endian_big;			/* byte order, 0 if little endian */
	int buf_ver;			/* LKST buffer version */
	char arch[LKST_ARCH_NAME_LEN];	/* Architecture name */
};

struct posix_log_entry {
	unsigned int log_magic;
	posix_log_recid_t log_recid;	/* ID of the event record */
	size_t log_size;	/* size of the event record variable data */
	int log_format;		/* format of variable data */
	int log_event_type;	/* event identification code */
	posix_log_facility_t log_facility;	/* event facility code */
	posix_log_severity_t log_severity;	/* event severity code */
	uid_t log_uid;	/* effective user ID associated with the event */
	gid_t log_gid;	/* effective group ID associated with the event */
	pid_t log_pid;	/* process ID associated with event */
	pid_t log_pgrp;	/* process group associated with event */
	struct timespec log_time;	/* event time stamp */
	unsigned int log_flags;		/* bitmap of event flag */
	pthread_t log_thread;		/* thread ID associated with event */
	posix_log_procid_t log_processor; /* processor ID associated with event */
};

struct lkst_log_record {
	struct posix_log_entry posix;	/* log form specified by POSIX */
	lkst_arg_t log_arg1;
			/* 1st argument acquired at a trace point */
	lkst_arg_t log_arg2;
			/* 2nd argument acquired at a trace point */
	lkst_arg_t log_arg3;
			/* 3rd argument acquired at a trace point */
	lkst_arg_t log_arg4;
			/* 4th argument acquired at a trace point */
};

struct lkst_buffer_param {
	lkst_buffer_id id;	/* event buffer ID */
	int cpu;		/* cpu number */
	size_t size;		/* size of event buffer */
	size_t result_size;	/* result size of event buffer */
};

struct lkst_buffer_listent {
	size_t size;			/* buffer size */
	size_t readp;
	size_t writep;
	unsigned int wrap_flag;
	unsigned int current_flag;
	lkst_buffer_id id, prev, next;	/* buffer ID of own/prev/next buffer */
	int cpu;			/* cpu number of the buffer */
};

#if DEBUG
struct lkst_buffer_seekparam {
	int offset;		/* offset of a read pointer */
	lkst_buffer_id id;	/* event buffer ID */
	size_t result_readp;	/* changed read pointer */
};
#endif /* DEBUG */

struct lkst_buffer_listparam {
	size_t listent_size;		/* size of the listent */
	struct lkst_buffer_listent *listent;
				/* area to store the list of event buffers */
};

struct lkst_buffer_ringparam {
	int cpu;	/* cpu number */
	int ring;	/* ring flag */
};

struct lkst_buffer_srmodparam {
	int cpu;	/* cpu number */
	int mode;	/* read mode */
	struct timeval xtime;		/* xtime */
	lkst_tsc_t tsc;			/* machine cycle */ 
					/* These two will be used to calculate 
					 * time of events in real time. */
	lkst_cpu_freq_t	cpu_freq;	/* cpu clockspeed in kHz */ 
	int endian_big;			/* byte order, 0 if little endian */
	int buf_ver;			/* LKST buffer version */
	char arch[LKST_ARCH_NAME_LEN];	/* Architecture name */
};

int lkst_buffer_read(struct lkst_log_buffer *lbuffer);

int lkst_buffer_create(struct lkst_buffer_param *buffer_param);

int lkst_buffer_shift(int cpu);

int lkst_buffer_delete(lkst_buffer_id id);

#if DEBUG
int lkst_buffer_seek(struct lkst_buffer_seekparam *bs);
#endif /* DEBUG */

int lkst_buffer_list(struct lkst_buffer_listparam *lp);

int lkst_buffer_ring(struct lkst_buffer_ringparam *rp);

int lkst_buffer_setrmod(struct lkst_buffer_srmodparam *sp);

#define LKST_BUFFER_LISTENT_SIZE(n)					\
	     (sizeof(struct lkst_buffer_listent) * (n))

#define LKST_BUFFER_SIZE_LOCAL2EXT(size) 				\
		((size)*sizeof(struct lkst_log_record)/LKST_SIZEOF_LKST_EVENT_RECORD)

#endif				/* _LINUX_LKST_BUFFER_H */

/* END */
