/*
 * linux/kernel/context.c
 *
 * Mechanism for running arbitrary tasks in process context
 *
 * dwmw2@redhat.com:		Genesis
 *
 * andrewm@uow.edu.au:		2.4.0-test12
 *	- Child reaping
 *	- Support for tasks which re-add themselves
 *	- flush_scheduled_tasks.
 *
 * arjanv@redhat.com:		
 *	- per cpu keventd
 */

#define __KERNEL_SYSCALLS__

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/init.h>
#include <linux/unistd.h>
#include <linux/signal.h>
#include <linux/interrupt.h>


struct kevent_info {
	struct list_head taskq;
	wait_queue_head_t waitq;
	struct task_struct *task;
} ____cacheline_aligned;


static struct kevent_info keventd[NR_CPUS];


static DECLARE_WAIT_QUEUE_HEAD(context_task_done);

/*
 * This function is needed by kmod for the following case:
 * a kevent is scheduled that exec's modprobe for a certain module.
 * that module invokes /sbin/hotplug via keventd and then waits for it
 * to complete. With this hack, kmod will exec hotplug directly instead
 * to avoid the recursion.
 */
int current_is_keventd(void)
{
	int i;
	for (i = 0; i < smp_num_cpus; i++)
		if (current == keventd[i].task)
			return 1;
	return 0;
}

	
/**
 * schedule_task - schedule a function for subsequent execution in process context.
 * @task: pointer to a &tq_struct which defines the function to be scheduled.
 *
 * May be called from interrupt context.  The scheduled function is run at some
 * time in the near future by the keventd kernel thread.  If it can sleep, it
 * should be designed to do so for the minimum possible time, as it will be
 * stalling all other scheduled tasks.
 *
 * schedule_task() returns non-zero if the task was successfully scheduled.
 * If @task is already residing on a task queue then schedule_task() fails
 * to schedule your task and returns zero.
 */
static int __schedule_task_cpu(struct tq_struct *task, int cpu)
{
	int ret;
	ret = queue_task(task, &keventd[cpu].taskq);
	wake_up(&keventd[cpu].waitq);
	return ret;
}

int schedule_task(struct tq_struct *task)
{
	return __schedule_task_cpu(task, cpu());
}

static int context_thread(void *dummy)
{
	struct task_struct *curtask = current;
	DECLARE_WAITQUEUE(wait, curtask);
	struct k_sigaction sa;
	int preferred_cpu = (long)dummy;

	daemonize();
	strcpy(current->comm, "keventd");

	keventd[preferred_cpu].task = curtask;

	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked, sigmask(SIGCHLD));
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);

	/* Install a handler so SIGCLD is delivered */
	sa.sa.sa_handler = SIG_IGN;
	sa.sa.sa_flags = 0;
	siginitset(&sa.sa.sa_mask, sigmask(SIGCHLD));
	do_sigaction(SIGCHLD, &sa, (struct k_sigaction *)0);

	
	set_cpus_allowed(current, 1UL << preferred_cpu);


	for (;;) {
		int queuenr = cpu();  
		set_current_state(TASK_INTERRUPTIBLE);
		add_wait_queue(&keventd[queuenr].waitq, &wait);	
		
		/*
		 * If one of the functions on a task queue re-adds itself
		 * to the task queue we call schedule() in state TASK_RUNNING
		 */
		if (TQ_ACTIVE(keventd[queuenr].taskq))
			set_current_state(TASK_RUNNING);
		schedule();
		remove_wait_queue(&keventd[queuenr].waitq, &wait);
	
		while (TQ_ACTIVE(keventd[queuenr].taskq)) 
			run_task_queue(&keventd[queuenr].taskq);
	
		wake_up(&context_task_done);
		if (signal_pending(current)) {
			while (waitpid(-1, (unsigned int *)0, __WALL|WNOHANG) > 0)
				;
			spin_lock_irq(&current->sigmask_lock);
			flush_signals(current);
			recalc_sigpending(current);
			spin_unlock_irq(&current->sigmask_lock);
		}
	}
}

/**
 * flush_scheduled_tasks - ensure that any scheduled tasks have run to completion.
 *
 * Forces execution of the schedule_task() queue and blocks until its completion.
 *
 * If a kernel subsystem uses schedule_task() and wishes to flush any pending
 * tasks, it should use this function.  This is typically used in driver shutdown
 * handlers.
 *
 * The caller should hold no spinlocks and should hold no semaphores which could
 * cause the scheduled tasks to block.
 */
static struct tq_struct dummy_task;

void flush_scheduled_tasks(void)
{
	int count, cpunr;
	DECLARE_WAITQUEUE(wait, current);

	/*
	 * Do it twice. It's possible, albeit highly unlikely, that
	 * the caller queued a task immediately before calling us,
	 * and that the eventd thread was already past the run_task_queue()
	 * but not yet into wake_up(), so it woke us up before completing
	 * the caller's queued task or our new dummy task.
	 *
	 * Also, do it for every cpu's queue since we have to flush them all out.
	 */
	add_wait_queue(&context_task_done, &wait);
	for (count = 0; count < 2; count++) {
		for (cpunr = 0; cpunr < smp_num_cpus; cpunr++) {
			set_current_state(TASK_UNINTERRUPTIBLE);

			/* Queue a dummy task to make sure we get kicked */
			__schedule_task_cpu(&dummy_task, cpunr);

			/* Wait for it to complete */
			schedule();
		}
	}
	remove_wait_queue(&context_task_done, &wait);
}
	
int start_context_thread(void)
{
	unsigned long i;
	for (i = 0; i < NR_CPUS; i++)
		INIT_LIST_HEAD(&keventd[i].taskq);
	
	for (i = 0; i < NR_CPUS; i++)
		init_waitqueue_head(&keventd[i].waitq);
	
	for (i = 0; i < smp_num_cpus; i++) {
		if (kernel_thread(context_thread, (void *) i,
				CLONE_FS | CLONE_FILES) < 0)
			printk("kernel_thread() for keventd failed for cpu %ld\n", i);
		sys_sched_yield();
	}
	return 0;
}

EXPORT_SYMBOL(schedule_task);
EXPORT_SYMBOL(flush_scheduled_tasks);

