/*****************************************************************************/
/* Kernel Hooks Interface.                                    		     */
/* Version 1.2                                                               */
/* Author: Richard J Moore richardj_moore@uk.ibm.com                         */
/* 	   Vamsi Krishna S. r1vamsi@in.ibm.com 				     */
/*                                                                           */
/* Implementation of the kernel hooks interface.			     */
/*                                                                           */
/* Refer to Documentation/hook/HOWTO for details.                            */
/*****************************************************************************/

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <linux/spinlock.h>
#include <linux/string.h>
#include <linux/proc_fs.h>
#include <linux/hook.h>

#ifndef FLASE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE !FALSE
#endif

#ifdef HOOK_DEBUG
#define DPRINTK(fmt, args...) printk(KERN_DEBUG "%s: "fmt, __FUNCTION__, ##args)
#else
#define DPRINTK(fmt, args...)
#endif

static spinlock_t hook_lock = SPIN_LOCK_UNLOCKED;

static int hook_global_status = 0;
static hook_head_t *last_hook_head;
static hook_head_t *first_hook_head;
static struct proc_dir_entry *proc_hooks_dir;
#define PROC_HOOKS_ROOT "hooks"

static int hook_default_exit(hook_rec_t * phookrec, unsigned int cParms, ...)
{
#ifdef HOOK_DEBUG
	int i;
	DPRINTK("cParms: %u pHook_record: 0x%08x\n", cParms, (unsigned int) phookrec);
	for(i = 1; i <= cParms; ++i)
		DPRINTK("Parm%u: %u\n", i, *(int *) *(&cParms + i));
#endif
	return HOOK_TERMINATE;
}

/* internal subroutines called by the public interfaces:
   hook_exit_register
   hook_exit_deregister
   hook_exit_arm
   hook_exit_disarm
   hook_initialise
   hook_terminate
*/

static void calculate_indicies(hook_head_t * phhr)
{
	hook_rec_t *phr;
 	/* recalculate indicies */
	for(phr = phhr->hook_next; phr != (hook_rec_t *) phhr; phr = phr->hook_next) {
		phr->hook_index = (phr->hook_prev->hook_index) + 1;
	}
	return;
}

static inline void deactivate_hook(hook_head_t * phookhead)
{
	phookhead->hook_flags &= ~HOOK_HEAD_ACTIVE;
	if(phookhead->hook_flags & HOOK_HEAD_ASM_HOOK)
		deactivate_asm_hook(phookhead);
}

static inline void activate_hook(hook_head_t * phookhead)
{
	phookhead->hook_flags |= HOOK_HEAD_ACTIVE;
	if(phookhead->hook_flags & HOOK_HEAD_ASM_HOOK)
		activate_asm_hook(phookhead);
}

static void disarm_one_hook_exit(hook_rec_t * phookrec)
{
	int found = FALSE;
	hook_rec_t *phr;

	phookrec->hook_flags &= ~HOOK_ARMED;

	/* see if there are any armed exits still present for this hook */
	for(phr = phookrec->hook_head->hook_next; phr != (hook_rec_t *) phookrec->hook_head;
	    phr = phr->hook_next) {

		if(phr->hook_flags & HOOK_ARMED) {
			found = TRUE;
			break;
		}
	}

	/* deactivate hook if no armed exits remain */
	if(!found) {
		deactivate_hook(phr->hook_head);
	}
	return;
}

/* validate input hook registeration record */
static int validate_input_hook_rec(hook_rec_t * phookrec)
{
        if(phookrec->hook_flags & ~HOOK_VALID_USER_FLAGS) {
		DPRINTK("Invalid hook_flags %08x\n", phookrec->hook_flags);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	if(!(phookrec->hook_exit)) {
		DPRINTK("Invalid hook_exit %08x\n", phookrec->hook_exit);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	if((phookrec->hook_next)) {
		DPRINTK("Invalid hook_nextflags %08x\n", phookrec->hook_next);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	if((phookrec->hook_prev)) {
		DPRINTK("Invalid hook_prev %08x\n", phookrec->hook_prev);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	if((phookrec->hook_index)) {
		DPRINTK("Invalid hook_index %08x\n", phookrec->hook_index);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	if((phookrec->hook_head)) {
		DPRINTK("Invalid hHead %08x\n", phookrec->hook_head);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	return 0;
}

/* validate a current hook registeration record*/
static int validate_hook_rec(hook_rec_t * phookrec)
{
	if(!(phookrec->hook_flags & HOOK_REGISTERED)) {
		DPRINTK("Invalid hook_flags %08x\n", phookrec->hook_flags);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	if(!(phookrec->hook_exit)) {
		DPRINTK("Invalid hook_exit %08x\n", phookrec->hook_exit);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	if(!(phookrec->hook_next)) {
		DPRINTK("Invalid hook_next %08x\n", phookrec->hook_next);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	if(!(phookrec->hook_prev)) {
		DPRINTK("Invalid hook_prev %08x\n", phookrec->hook_prev);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	if(!(phookrec->hook_index)) {
		DPRINTK("Invalid hook_index %08x\n", phookrec->hook_index);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	if(!(phookrec->hook_head)) {
		DPRINTK("Invalid hook_head %08x\n", phookrec->hook_head);
		return ERROR_INVALID_REGISTRATION_RECORD;
	}
	return 0;
}

/* common read/write routines for procfs entries */
static int hook_read_proc(char *page, char **start, off_t off, int count, 
		int *eof, void *data)
{
	int len;
	hook_rec_t *phookrec = (hook_rec_t *)data;
	len = sprintf(page, "%d\n", (phookrec->hook_flags & HOOK_ARMED));
	return len;
}

static int hook_write_proc(struct file *file, const char *buffer, 
		unsigned long count, void *data)
{
	hook_rec_t *phookrec = (hook_rec_t *)data;
	hook_rec_t *old_link = phookrec->hook_link;
	char input;
	int armed;

	if (copy_from_user(&input, buffer, 1)) {
		return -EFAULT;
	}
	armed = input - '0';
	if (armed && !(phookrec->hook_flags & HOOK_ARMED)) {
		hook_exit_arm(phookrec);
	} else if (!armed && (phookrec->hook_flags & HOOK_ARMED)) {
		hook_exit_disarm(phookrec);
	}
	phookrec->hook_link = old_link;
	return count; /* dont want to read anymore */
}

static int hook_create_proc(hook_rec_t * phookrec, int proc_writable)
{
	char *proc_entry_name;
	hook_head_t *phhr = (hook_head_t *) (phookrec->hook_head);

	if (phhr->proc_entry) {
		char tmp[32];
		proc_entry_name = phookrec->hook_exit_name;
		if (!proc_entry_name) {
			sprintf(tmp, "%p", phookrec);
			proc_entry_name = tmp;
		}
		if (!proc_writable) {
			phookrec->proc_entry = create_proc_read_entry(
					proc_entry_name, 0444, 
					phhr->proc_entry, 
					hook_read_proc,
					phookrec);
		} else {
			phookrec->proc_entry = create_proc_entry(
					proc_entry_name, 0644, 
					phhr->proc_entry);
			if (phookrec->proc_entry) {
				phookrec->proc_entry->data = phookrec;
				phookrec->proc_entry->read_proc = hook_read_proc;
				phookrec->proc_entry->write_proc = hook_write_proc;
			}
		}
	}
}

/* public interfaces being here ... */

int _hook_exit_register(char *hookid, hook_rec_t * phookrec, int proc_writable)
{
	hook_head_t *phhr;
	hook_rec_t *phr;
	int found;
	int rc;
	unsigned long eflags;

	if(!(hook_global_status & HOOK_INIT))
		return ERROR_GKHI_UNAVAILABLE;

        rc = validate_input_hook_rec(phookrec);
	if (rc)
		return rc;

	/* from this point we depend on GKHI structure consistency */
	spin_lock_irqsave(&hook_lock, eflags);

	/* locate the corresponding hookhead */
	found = FALSE;
	for(phhr = first_hook_head; phhr; phhr = phhr->hook_head_next) {
		if(strcmp(hookid, phhr->hook_id) == 0) {
			found = TRUE;
			break;
		}
	}

	if(!found) {
		spin_unlock_irqrestore(&hook_lock, eflags);
		return ERROR_HOOK_NOT_FOUND;
	}

	/* check for priority special conditions and determine the insertion
	 * point. Note if scheduling priority is not specified then exits are
	 * dispatched in FIFO order unless the LIFO flag is set, in which
	 * case the hook record is inseted at the head of the queue.
	 */
	phr = (hook_rec_t *) phhr;

	/*
	 * If the hook head is defined as an exclusive hook, force
	 * the hookrec flags to be exclusive.
	 */
	if(phhr->hook_flags & HOOK_HEAD_EXCLUSIVE) {
		phookrec->hook_flags |= HOOK_PRIORITY_EXCLUSIVE;
	}
	
	if((phookrec->hook_flags) & HOOK_PRIORITY_FIRST) {
		if(phhr->hook_next != (hook_rec_t *) phhr) {
			if((phhr->hook_next->hook_flags) & HOOK_PRIORITY_FIRST) {
				spin_unlock_irqrestore(&hook_lock, eflags);
				return ERROR_HIGHER_PRIORITY_HOOK;
			}
		}
	}

	if((phookrec->hook_flags) & HOOK_PRIORITY_LAST) {
		if(phhr->hook_next != (hook_rec_t *) phhr) {
			if((phhr->hook_prev->hook_flags) & HOOK_PRIORITY_LAST) {
				spin_unlock_irqrestore(&hook_lock, eflags);
				return ERROR_LOWER_PRIORITY_HOOK;
			} else {
				phr = phhr->hook_prev;
			}
		}
	}

	if(!((phookrec->hook_flags) & HOOK_PRIORITY)) {
		if((phookrec->hook_flags) & HOOK_QUEUE_LIFO) {
			if((phhr->hook_next->hook_flags) & HOOK_PRIORITY_FIRST)
				phr = phhr->hook_next;
		} else if((phhr->hook_prev->hook_flags) & HOOK_PRIORITY_LAST)
			phr = phhr->hook_prev->hook_prev;
		else
			phr = phhr->hook_prev;

	}

	/*
	 * phr points to the hook record after which we make the insertion -
	 * insert into circular double-linked list
	 */
	phookrec->hook_next = phr->hook_next;
	phr->hook_next = phookrec;
	phookrec->hook_prev = phr;
	phookrec->hook_next->hook_prev = phookrec;

	/* set output fields */

	/* link back to head */
	(hook_head_t *) (phookrec->hook_head) = phhr;

	/* recalculate indicies */
	calculate_indicies(phookrec->hook_head);

	/* 
	 * create proc entry if exit name is specified 
	 * and a proc entry exists for the hook head exits. 
	 */
	if(!(hook_global_status & HOOK_IN_START)) {
		hook_create_proc(phookrec, proc_writable);
	}
	else {
		phookrec->proc_writable = proc_writable;
	}

	/* optimization for exclusive hooks */
	if (phhr->hook_flags & HOOK_HEAD_EXCLUSIVE) {
		phhr->hook_exit = phookrec->hook_exit;
	}

	/* indicate registration compete */
	phookrec->hook_flags |= HOOK_REGISTERED;

	spin_unlock_irqrestore(&hook_lock, eflags);
	return NO_ERROR;
}

int hook_exit_deregister(hook_rec_t * phookrec)
{
	hook_rec_t *phr;
	int rc;
	unsigned long eflags;
	char *proc_entry_name;

	if(!(hook_global_status & HOOK_INIT))
		return ERROR_GKHI_UNAVAILABLE;

	/* from this point we depend on GKHI structure consistency */
	spin_lock_irqsave(&hook_lock, eflags);
	for(phr = phookrec; phr; phr = phr->hook_link) {
		rc = validate_hook_rec(phr);
		if (rc) {
			spin_unlock_irqrestore(&hook_lock, eflags);
			return rc;
		}

		/* disarm if armed */
		if(phr->hook_flags & HOOK_ARMED)
			disarm_one_hook_exit(phr);

		/* unlink from chain head */
		phr->hook_prev->hook_next = phr->hook_next;
		phr->hook_next->hook_prev = phr->hook_prev;

		/* re-calculate indicies.
		 * Note this is possibly inefficient to do this here if more than
		 * one exit is deregistered from a single hook. But that should be
		 * unlikely since a kernel module would normaly register or deregister
		 * one exit per hook.
		 */
		calculate_indicies(phr->hook_head);

		/* cleanup proc entry, if any */
		if (phr->proc_entry) {
			char tmp[32];
			proc_entry_name = phr->hook_exit_name;
			if (!proc_entry_name) {
				sprintf(tmp, "%p", phookrec);
				proc_entry_name = tmp;
			}
			phr->proc_entry->data = NULL;
			remove_proc_entry(proc_entry_name, phr->hook_head->proc_entry);
		}

		/* optimization for exclusive hooks */
		if (phr->hook_head->hook_flags & HOOK_HEAD_EXCLUSIVE) {
			phr->hook_head->hook_exit = &hook_default_exit;
		}

		/* indicate deregistration compete */
		phr->hook_flags &= ~HOOK_REGISTERED;

	}
	spin_unlock_irqrestore(&hook_lock, eflags);
	return NO_ERROR;
}

int hook_exit_arm(hook_rec_t * phookrec)
{
	hook_rec_t *phr;
	int rc;
	unsigned long eflags;

	if(!(hook_global_status & HOOK_INIT))
		return ERROR_GKHI_UNAVAILABLE;

	/* from this point we depend on GKHI structure consistency */
	spin_lock_irqsave(&hook_lock, eflags);
	for(phr = phookrec; phr; phr = phr->hook_link) {
		rc = validate_hook_rec(phr);
		if (rc) {
			spin_unlock_irqrestore(&hook_lock, eflags);
			return rc;
		}

		/* mark armed */
		phr->hook_flags |= HOOK_ARMED;

		/* activate hook if not already active */
		if(!(phr->hook_head->hook_flags & HOOK_HEAD_ACTIVE)) {
			activate_hook(phr->hook_head);
		}

	}
	spin_unlock_irqrestore(&hook_lock, eflags);
	return NO_ERROR;
}

int hook_exit_disarm(hook_rec_t * phookrec)
{
	hook_rec_t *phr;
	int rc;
	unsigned long eflags;

	if(!(hook_global_status & HOOK_INIT))
		return ERROR_GKHI_UNAVAILABLE;

	/* from this point we depend on GKHI structure consistency */
	spin_lock_irqsave(&hook_lock, eflags);
	for(phr = phookrec; phr; phr = phr->hook_link) {
		rc = validate_hook_rec(phr);
		if (rc) {
			spin_unlock_irqrestore(&hook_lock, eflags);
			return rc;
		}

		/* make sure that it is armed */
		if(!(phr->hook_flags & HOOK_ARMED)) {
			spin_unlock_irqrestore(&hook_lock, eflags);
			DPRINTK("Invalid hook_flags %08x\n", phookrec->hook_flags);
			return ERROR_INVALID_REGISTRATION_RECORD;
		}
		disarm_one_hook_exit(phr);
	}
	spin_unlock_irqrestore(&hook_lock, eflags);
	return NO_ERROR;
}

int _hook_initialise(hook_head_t * phookhead)
{
	hook_head_t *phhr;
	unsigned char *phookloc = (unsigned char *)(phookhead->hook_loc);
	int found;
	int rc;
	unsigned long eflags;

	if(!(hook_global_status & HOOK_INIT))
		return ERROR_GKHI_UNAVAILABLE;

	rc = is_asm_hook(phookloc);
	if(rc) {
	       if(rc == 1) 
			phookhead->hook_flags |= HOOK_HEAD_ASM_HOOK;
		else
			return rc;
	}

	/* from this point we depend on GKHI structure consistency */
	spin_lock_irqsave(&hook_lock, eflags);

	/* ensure hookid is not in use */
	found = FALSE;
	for(phhr = first_hook_head; phhr; phhr = phhr->hook_head_next) {
		if(strcmp(phookhead->hook_id, phhr->hook_id) == 0) {
			found = TRUE;
			break;
		}
	}

	if(found) {
		spin_unlock_irqrestore(&hook_lock, eflags);
		return ERROR_HOOK_EXISTS;
	}

	/* link and initialise new hookhead */
	phookhead->hook_exit = &hook_default_exit;
	phookhead->hook_next = (hook_rec_t *) phookhead;
	phookhead->hook_prev = (hook_rec_t *) phookhead;
	phookhead->hook_flags |= HOOK_ARMED;
	phookhead->hook_link = NULL;
	phookhead->hook_head = phookhead;
	phookhead->hook_index = 0;
	phookhead->hook_head_next = NULL;
	phookhead->hook_loc = phookloc;

	if(!(hook_global_status & HOOK_IN_START)) {
		phookhead->proc_entry = proc_mkdir(phookhead->hook_id, proc_hooks_dir);
	}

	/* chain to last hookhead or initialise first and last hookhead pointers */
	if(last_hook_head)
		last_hook_head->hook_head_next = phookhead;
	else
		first_hook_head = phookhead;
	last_hook_head = phookhead;

	spin_unlock_irqrestore(&hook_lock, eflags);
	return NO_ERROR;
}

int _hook_terminate(hook_head_t * phookhead, int force)
{
	hook_head_t *phhr;
	hook_head_t *pprevhhr;
	int found;
	unsigned long eflags;

	if(!(hook_global_status & HOOK_INIT))
		return ERROR_GKHI_UNAVAILABLE;

	/* from this point we depend on GKHI structure consistency */
	spin_lock_irqsave(&hook_lock, eflags);

	/* verify hookhead structure is in the chain of hookheads */
	found = FALSE;
	pprevhhr = NULL;
	for(phhr = first_hook_head; phhr; phhr = phhr->hook_head_next) {
		if(phhr == phookhead) {
			found = TRUE;
			break;
		}
		pprevhhr = phhr;
	}

	if(!found) {
		spin_unlock_irqrestore(&hook_lock, eflags);
		return ERROR_HOOK_NOT_FOUND;
	}

	if(phhr->hook_flags & HOOK_HEAD_ACTIVE) {
		if(force) {
			deactivate_hook(phhr);
		} else {
			spin_unlock_irqrestore(&hook_lock, eflags);
			return ERROR_HOOK_IS_ACTIVE;
		}
	}

	if (phhr->proc_entry) {
		remove_proc_entry(phhr->hook_id, proc_hooks_dir);
	}

	/* remove from chain */
	if(pprevhhr)
		pprevhhr->hook_head_next = phhr->hook_head_next;
	if(phhr == last_hook_head)
		last_hook_head = pprevhhr;
	if(phhr == first_hook_head)
		first_hook_head = phhr->hook_head_next;

	spin_unlock_irqrestore(&hook_lock, eflags);
	return NO_ERROR;
}

int hook_init_in_start(void)
{
	hook_global_status = 0;
	hook_global_status |= HOOK_INIT;
	hook_global_status |= HOOK_IN_START;
	/* @@: for kernel hooks, add hook_initialise statement here. */

	printk(KERN_INFO "Generalised Kernel Hooks Interface 1.4\n");
	DPRINTK("Debug version active\n");
	return 0;
}

static int __init hook_init_module(void)
{
	hook_head_t *phookhead;
	hook_rec_t *phookrec;
	unsigned long eflags;
	int ret = 0;
	if(!(hook_global_status & HOOK_IN_START)) {
		spin_lock_irqsave(&hook_lock, eflags);
		proc_hooks_dir = proc_mkdir(PROC_HOOKS_ROOT, NULL);
		ret = hook_init_in_start();
		spin_unlock_irqrestore(&hook_lock, eflags);
	}
	else {
		spin_lock_irqsave(&hook_lock, eflags);
		hook_global_status &= ~HOOK_IN_START;

		/* Hooks has been already initialized. Now make proc entry. */
		proc_hooks_dir = proc_mkdir(PROC_HOOKS_ROOT, NULL);
		for(phookhead = first_hook_head; phookhead; phookhead = (phookhead->hook_head_next)) {
			phookhead->proc_entry = proc_mkdir(phookhead->hook_id, proc_hooks_dir);
			for(phookrec = phookhead->hook_next;
			    phookrec != (hook_rec_t *) phookhead; phookrec = phookrec->hook_next) {
					hook_create_proc(phookrec, phookrec->proc_writable);
			}
		}
		spin_unlock_irqrestore(&hook_lock, eflags);
	}
	return ret;
}

static void /*__exit*/ hook_cleanup_module(void)
{
	hook_head_t *phookhead;
	hook_rec_t *phookrec;
	unsigned long eflags;

	spin_lock_irqsave(&hook_lock, eflags);

	/* deactivate each hook */
	for(phookhead = first_hook_head; phookhead; phookhead = (phookhead->hook_head_next)) {
		/* disarm and de-register each hook exit */
		for(phookrec = phookhead->hook_next;
		    phookrec != (hook_rec_t *) phookhead; phookrec = phookrec->hook_next) {
        			phookrec->hook_flags &= ~(HOOK_REGISTERED | HOOK_ARMED);
		}
		if(phookhead->hook_flags & HOOK_HEAD_ACTIVE) {
			deactivate_hook(phookhead);
		}
	}

	remove_proc_entry(PROC_HOOKS_ROOT, NULL);

	spin_unlock_irqrestore(&hook_lock, eflags);
	printk(KERN_INFO "Generalised Kernel Hooks Interface 1.4 terminated\n");
	return;
}

module_init(hook_init_module);
module_exit(hook_cleanup_module);

EXPORT_SYMBOL(hook_exit_deregister);
EXPORT_SYMBOL(hook_exit_arm);
EXPORT_SYMBOL(hook_exit_disarm);
EXPORT_SYMBOL(_hook_initialise);
EXPORT_SYMBOL(_hook_terminate);
EXPORT_SYMBOL(_hook_exit_register);
EXPORT_SYMBOL(hook_init_in_start);
