/*
 *  linux/mm/vmscan.c
 *
 *  Copyright (C) 1991, 1992, 1993, 1994  Linus Torvalds
 *
 *  Swap reorganised 29.12.95, Stephen Tweedie.
 *  kswapd added: 7.1.96  sct
 *  Removed kswapd_ctl limits, and swap out as many pages as needed
 *  to bring the system back to freepages.high: 2.4.97, Rik van Riel.
 *  Version: $Id: vmscan.c,v 1.5 1998/02/23 22:14:28 sct Exp $
 *  Zone aware kswapd started 02/00, Kanoj Sarcar (kanoj@sgi.com).
 *  Multiqueue VM started 5.8.00, Rik van Riel.
 */

#include <linux/slab.h>
#include <linux/kernel_stat.h>
#include <linux/swap.h>
#include <linux/swapctl.h>
#include <linux/smp_lock.h>
#include <linux/pagemap.h>
#include <linux/init.h>
#include <linux/highmem.h>
#include <linux/file.h>

#include <asm/pgalloc.h>

int vm_static_inactive_target;
int total_reserved_pages = 0;
int vmfixes = 1;

#ifndef min
#define min(a,b) (((a)<(b))?(a):(b))
#endif
#ifndef max
#define max(a,b) (((a)>(b))?(a):(b))
#endif

#define DEF_PRIORITY 6

struct age_parms_s vm_age_parms = {
	PAGE_AGE_START,
	PAGE_AGE_ADV, 
	PAGE_AGE_DECL, 
	PAGE_AGE_MAX
};

static inline void age_page_up(struct page *page)
{
	page->age = min((int) (page->age + vm_age_parms.adv), vm_age_parms.max); 
}

static inline void age_page_down(struct page *page)
{
	page->age -= min(vm_age_parms.decl, (int)page->age);
}

#include <linux/lkst.h>

/*
 * Estimate whether a zone has enough inactive or free pages..
 */
static unsigned int zone_inactive_plenty(zone_t *zone)
{
	unsigned int inactive;

	if (!zone->size)
		return 0;
		
	inactive = zone->inactive_dirty_pages;
	inactive += zone->inactive_clean_pages;
	inactive += zone->free_pages;

	return (inactive > (zone->size / 3));
}

static unsigned int zone_free_plenty(zone_t *zone)
{
	unsigned int free;

	free = zone->free_pages;
	free += zone->inactive_clean_pages;

	return free > zone->pages_high*2;
}

/*
 * In general, page aging can balance the various uses of memory
 * nicely, but sometimes the caches have so much activity that
 * they push out other things and influence system behaviour in
 * a bad way.  If we have too much of a certain cache, we just
 * bypass page aging and drop cache pages a bit faster.
 */
static inline int too_many_buffers(void)
{
	int bufferpages = atomic_read(&buffermem_pages);
	int limit = num_physpages * buffer_mem.borrow_percent / 100;

	return bufferpages > limit;
}

static inline int pagecache_too_large(void)
{
	int pagecache = atomic_read(&page_cache_size) - swapper_space.nrpages;
	int limit = num_physpages * page_cache.borrow_percent / 100;

	return pagecache > limit;
}

/*
 * The swap-out function returns 1 if it successfully
 * scanned all the pages it was asked to (`count').
 * It returns zero if it couldn't do anything,
 *
 * rss may decrease because pages are shared, but this
 * doesn't count as having freed a page.
 */

/* mm->page_table_lock is held. mmap_sem is not held */
static int try_to_swap_out(struct mm_struct * mm, struct vm_area_struct* vma, unsigned long address, pte_t * page_table, struct page *page)
{
	pte_t pte;
	swp_entry_t entry;

	/* Don't look at this pte if it's been accessed recently. */
	if (ptep_test_and_clear_young(page_table)) {
		age_page_up(page);
		return 1;
	}


	/* 
	 * If we have plenty inactive pages on this 
	 * zone, skip it.
	 */
	if (zone_inactive_plenty(page->zone))
		return 1;

	/*
	 * If the page is on the active list, page aging is done in
	 * refill_inactive_scan(), anonymous pages are aged here.
	 * This is done so heavily shared pages (think libc.so)
	 * don't get punished heavily while they are still in use.
	 * The alternative would be to put anonymous pages on the
	 * active list too, but that increases complexity (for now).
	 */
	if (!PageActive(page))
		age_page_down(page);

	/*
	 * Don't swap out a page which is still in use.
	 */
	if (page->age > 0)
		return 1;

	if (TryLockPage(page))
		return 1;

	/* From this point on, the odds are that we're going to
	 * nuke this pte, so read and clear the pte.  This hook
	 * is needed on CPUs which update the accessed and dirty
	 * bits in hardware.
	 */
	flush_cache_page(vma, address);
	pte = vm_ptep_get_and_clear(vma, page_table, address);
	flush_tlb_page(vma, address);

	/*
	 * Is the page already in the swap cache? If so, then
	 * we can just drop our reference to it without doing
	 * any IO - it's already up-to-date on disk.
	 */
	if (PageSwapCache(page)) {
		entry.val = page->index;
		if (pte_dirty(pte))
			set_page_dirty(page);
set_swap_pte:
		swap_duplicate(entry);
		vm_set_pte(vma, page_table, swp_entry_to_pte(entry), address);
drop_pte:
		mm->rss--;
		if (!page->age)
			deactivate_page(page);
		UnlockPage(page);
		page_cache_release(page);
		return 1;
	}

	/*
	 * Is it a clean page? Then it must be recoverable
	 * by just paging it in again, and we can just drop
	 * it..  or if it's dirty but has backing store,
	 * just mark the page dirty and drop it.
	 *
	 * However, this won't actually free any real
	 * memory, as the page will just be in the page cache
	 * somewhere, and as such we should just continue
	 * our scan.
	 *
	 * Basically, this just makes it possible for us to do
	 * some real work in the future in "refill_inactive()".
	 */
	if (page->mapping) {
		if (pte_dirty(pte))
			set_page_dirty(page);
		goto drop_pte;
	}
	/*
	 * Check PageDirty as well as pte_dirty: page may
	 * have been brought back from swap by swapoff.
	 */
	if (!pte_dirty(pte) && !PageDirty(page))
		goto drop_pte;

	/*
	 * This is a dirty, swappable page.  First of all,
	 * get a suitable swap entry for it, and make sure
	 * we have the swap cache set up to associate the
	 * page with that swap entry.
	 */
	entry = get_swap_page();
	if (!entry.val)
		goto out_unlock_restore; /* No swap space left */

	/* Add it to the swap cache and mark it dirty */
	add_to_swap_cache(page, entry);
	set_page_dirty(page);
	goto set_swap_pte;

out_unlock_restore:
	vm_set_pte(vma, page_table, pte, address);
	UnlockPage(page);
	return 0;			/* no swap space */
}

/* mm->page_table_lock is held. mmap_sem is not held */
static int swap_out_pmd(struct mm_struct * mm, struct vm_area_struct * vma, pmd_t *dir, unsigned long address, unsigned long end, int count)
{
	unsigned long pmd_end;
	pte_t *pte;
	int made_progress = 0;

	if (pmd_none(*dir))
		return count;
	if (pmd_bad(*dir)) {
		pmd_ERROR(*dir);
		pmd_clear(dir);
		return count;
	}
	
	pte = pte_offset_map(dir, address);
	
	pmd_end = (address + PMD_SIZE) & PMD_MASK;
	if (end > pmd_end)
		end = pmd_end;

	do {
		if (pte_present(*pte)) {
			struct page *page = pte_page(*pte);

			if (VALID_PAGE(page) && !PageReserved(page)) {
				if (!try_to_swap_out(mm, vma, address, pte, page) && vmfixes) {
					mm->swap_address = address;
					pte_unmap(pte);
					return 0;
				}
				made_progress++;
				if (!--count) {
					pte++;
					break;
				}
			}
		}
		address += PAGE_SIZE;
		pte++;
	} while (address && (address < end));
	if (vmfixes && !made_progress) count--;
	mm->swap_address = address + PAGE_SIZE;
	pte_unmap(pte - 1);
	return count;
}

/* mm->page_table_lock is held. mmap_sem is not held */
static inline int swap_out_pgd( struct mm_struct * mm, struct vm_area_struct * vma, pgd_t *dir, unsigned long address, unsigned long end, int count)
{
	pmd_t * pmd;
	unsigned long pgd_end;

	if (pgd_none(*dir))
		return count;
	if (pgd_bad(*dir)) {
		pgd_ERROR(*dir);
		pgd_clear(dir);
		return count;
	}

	pmd = pmd_offset(dir, address);

	pgd_end = (address + PGDIR_SIZE) & PGDIR_MASK;	
	if (pgd_end && (end > pgd_end))
		end = pgd_end;
	
	do {
		count = swap_out_pmd(mm, vma, pmd, address, end, count);
		if (!count)
			break;
		address = (address + PMD_SIZE) & PMD_MASK;
		pmd++;
	} while (address && (address < end));
	return count;
}

/* mm->page_table_lock is held. mmap_sem is not held */
static int swap_out_vma(struct mm_struct * mm, struct vm_area_struct * vma, unsigned long address, int count)
{
	pgd_t *pgdir;
	unsigned long end;

	/* Don't swap out areas which are locked down */
	if (vma->vm_flags & (VM_LOCKED|VM_RESERVED))
		return count;

	pgdir = pgd_offset(mm, address);

	end = vma->vm_end;
	if (address >= end)
		BUG();
	do {
		count = swap_out_pgd(mm, vma, pgdir, address, end, count);
		if (!count)
			break;
		address = (address + PGDIR_SIZE) & PGDIR_MASK;
		pgdir++;
	} while (address && (address < end));
	return count;
}

/*
 * Returns non-zero if we scanned all `count' pages
 */
static int swap_out_mm(struct mm_struct * mm, int count)
{
	unsigned long address;
	struct vm_area_struct* vma;

	if (!count)
		return 1;
	/*
	 * Go through process' page directory.
	 */

	/*
	 * Find the proper vm-area after freezing the vma chain 
	 * and ptes.
	 */
	spin_lock(&mm->page_table_lock);
	address = mm->swap_address;
	vma = find_vma(mm, address);
	if (vma) {
		if (address < vma->vm_start)
			address = vma->vm_start;

		for (;;) {
			count = swap_out_vma(mm, vma, address, count);
			if (!count)
				goto out_unlock;
			vma = vma->vm_next;
			if (!vma)
				break;
			address = vma->vm_start;
		}
	}
	/* Reset to 0 when we reach the end of address space */
	mm->swap_address = 0;

out_unlock:
	spin_unlock(&mm->page_table_lock);
	return !count;
}

#define SWAP_SHIFT	5
#define SWAP_MIN	8

static inline int swap_amount(struct mm_struct *mm)
{
	int nr = mm->rss >> SWAP_SHIFT;
	if (nr < SWAP_MIN) {
		nr = SWAP_MIN;
		if (nr > mm->rss)
			nr = mm->rss;
	}
	return nr;
}

/* Placeholder for swap_out(): may be updated by fork.c:mmput() */
struct mm_struct *swap_mm = &init_mm;

static void swap_out(unsigned int priority, int gfp_mask)
{
	int counter;
	int retval = 0;
	struct mm_struct *mm = current->mm;

	/* Always start by trying to penalize the process that is allocating memory */
	if (mm)
		retval = swap_out_mm(mm, swap_amount(mm));

	/* Then, look at the other mm's */
	counter = max(1, (mmlist_nr << SWAP_SHIFT) >> priority);
	do {
		spin_lock(&mmlist_lock);
		mm = swap_mm;
		if (mm == &init_mm) {
			mm = list_entry(mm->mmlist.next, struct mm_struct, mmlist);
			if (mm == &init_mm)
				goto empty;
		}
		/* Set pointer for next call to next in the list */
		swap_mm = list_entry(mm->mmlist.next, struct mm_struct, mmlist);

		/* Make sure the mm doesn't disappear when we drop the lock.. */
		atomic_inc(&mm->mm_users);
		spin_unlock(&mmlist_lock);

		/* Walk about 6% of the address space each time */
		retval |= swap_out_mm(mm, swap_amount(mm));
		mmput(mm);
	} while (--counter >= 0);
	return;

empty:
	spin_unlock(&mmlist_lock);
}


/**
 * reclaim_page -	reclaims one page from the inactive_clean list
 * @zone: reclaim a page from this zone
 *
 * The pages on the inactive_clean can be instantly reclaimed.
 * The tests look impressive, but most of the time we'll grab
 * the first page of the list and exit successfully.
 */
struct page * reclaim_page(zone_t * zone)
{
	struct address_space *mapping;
	struct page * page = NULL;
	struct list_head * page_lru;
	spinlock_t *pg_lock;
	int maxscan;

	/* The repeating is necessary because we are taking these
	 * locks out of the normal order.
	 */
repeat:
	spin_lock(&pagemap_lru_lock);
	maxscan = zone->inactive_clean_pages;
	while ((page_lru = zone->inactive_clean_list.prev) !=
			&zone->inactive_clean_list && maxscan--) {
			
		page = list_entry(page_lru, struct page, lru);

		if (!page->mapping)
			goto refile_dirty;

		pg_lock = PAGECACHE_LOCK(page);
		if (!spin_trylock(pg_lock)) {
			spin_unlock(&pagemap_lru_lock);
			goto repeat;
		}

		/* Wrong page on list?! (list corruption, should not happen) */
		if (!PageInactiveClean(page)) {
			printk("VM: reclaim_page, wrong page on list.\n");
			list_del(page_lru);
			page->zone->inactive_clean_pages--;
			spin_unlock(pg_lock);
			continue;
		}

		/* Page is or was in use?  Move it to the active list. */
		if (PageReferenced(page) || page->age > 0 ||
				page_count(page) > (1 + !!page->buffers)) {
			spin_unlock(pg_lock);
refile_dirty:
			del_page_from_inactive_clean_list(page);
			add_page_to_active_list(page);
			continue;
		}

		/* The page is dirty, or locked, move to inactive_dirty list. */
		if (page->buffers || PageDirty(page) || TryLockPage(page)) {
			del_page_from_inactive_clean_list(page);
			add_page_to_inactive_dirty_list(page);
			spin_unlock(pg_lock);
			continue;
		}

		/* OK, remove the page from the caches. */
                if (PageSwapCache(page)) {
			LKST_HOOK_INLINE(LKST_ETYPE_MEM_SWAPOUT,
					 LKST_ARG32(page, 0),
					 LKST_ARG32(0, 0),
					 LKST_ARG32(0, 0),
					 LKST_ARG32(0, 0));
			__delete_from_swap_cache(page);
			goto found_page;
		}

		mapping = page->mapping;
		if (mapping) {
			spin_lock(&mapping->page_lock);
			__remove_inode_page(page);
			spin_unlock(&mapping->page_lock);
			goto found_page;
		}

		/* We should never ever get here. */
		BUG();
	}
	/* Reset page pointer, maybe we encountered an unfreeable page. */
	page = NULL;
	goto out;

found_page:
	spin_unlock(pg_lock);
	del_page_from_inactive_clean_list(page);
	UnlockPage(page);
	page->age = vm_age_parms.start;
	if (page_count(page) != 1)
		printk("VM: reclaim_page, found page with count %d!\n",
				page_count(page));

out:
	spin_unlock(&pagemap_lru_lock);
	return page;
}

static inline int page_dirty(struct page *page)
{
	struct buffer_head *tmp, *bh;

	if (PageDirty(page))
		return 1;

	if (page->mapping && !page->buffers)
		return 0;

	tmp = bh = page->buffers;

	do {
		if (tmp->b_state & ((1<<BH_Dirty) | (1<<BH_Lock)))
			return 1;
		tmp = tmp->b_this_page;
	} while (tmp != bh);

	return 0;
}

/**
 * page_launder - clean dirty inactive pages, move to inactive_clean list
 * @gfp_mask: what operations we are allowed to do
 * @sync: are we allowed to do synchronous IO in emergencies ?
 *
 * When this function is called, we are most likely low on free +
 * inactive_clean pages. Since we want to refill those pages as
 * soon as possible, we'll make two loops over the inactive list,
 * one to move the already cleaned pages to the inactive_clean lists
 * and one to (often asynchronously) clean the dirty inactive pages.
 *
 * In situations where kswapd cannot keep up, user processes will
 * end up calling this function. Since the user process needs to
 * have a page before it can continue with its allocation, we'll
 * do synchronous page flushing in that case.
 *
 * This code is heavily inspired by the FreeBSD source code. Thanks
 * go out to Matthew Dillon.
 */
#define MAX_LAUNDER 		(4 * (1 << page_cluster))
#define CAN_DO_FS		(gfp_mask & __GFP_FS)
#define CAN_DO_IO		(gfp_mask & __GFP_IO)
static int page_launder_zone(zone_t * zone, int gfp_mask, int sync, int full_flush)
{
	int launder_loop, maxscan, cleaned_pages, maxlaunder;
	struct list_head * page_lru;
	struct page * page;
	int want_bdflush = 0;
	extern int high_queued_sectors;
	extern atomic_t queued_sectors;

	launder_loop = 0;
	maxlaunder = 0;
	cleaned_pages = 0;

dirty_page_rescan:
	spin_lock(&pagemap_lru_lock);
	maxscan = zone->inactive_dirty_pages;
	
	while ((page_lru = zone->inactive_dirty_list.prev) != &zone->inactive_dirty_list &&
				maxscan-- > 0) {
				
		if (current->need_resched) {
			spin_unlock(&pagemap_lru_lock);
			schedule();
			spin_lock(&pagemap_lru_lock);			        
			continue;
		}
		page = list_entry(page_lru, struct page, lru);
		prefetch(page_lru->prev);

		/* Wrong page on list?! (list corruption, should not happen) */
		if (!PageInactiveDirty(page)) {
			printk("VM: page_launder, wrong page on list.\n");
			list_del(page_lru);
			nr_inactive_dirty_pages--;
			page->zone->inactive_dirty_pages--;
			continue;
		}

		/*
		 * If the page is or was in use, we move it to the active
		 * list. Note that pages in use by processes can easily
		 * end up here to be unmapped later, but we just don't
		 * want them clogging up the inactive list.
		 */
		if ((PageReferenced(page) || page->age > 0 ||
				page_count(page) > (1 + !!page->buffers) ||
				page_ramdisk(page)) && !PageLocked(page)) {
			del_page_from_inactive_dirty_list(page);
			add_page_to_active_list(page);
			page->age = max((int)page->age, vm_age_parms.start);
			continue;
		}

		/*
		 * The page is locked. IO in progress?
		 * Move it to the back of the list.
		 */
		if (TryLockPage(page)) {
			list_del(page_lru);
			list_add(page_lru, &zone->inactive_dirty_list);
			continue;
		}

		/*
		 * If we have plenty free pages on a zone: 
		 *
		 * 1) we avoid a writeout for that page if its dirty.
		 * 2) if its a buffercache page, and not a pagecache
		 * one, we skip it since we cannot move it to the 
		 * inactive clean list --- we have to free it.
		 */
		if (zone_free_plenty(page->zone)) {
			/* here be dragons! do not change this or it breaks */
			if (!page->mapping || page_dirty(page)) {
				list_del(page_lru);
				list_add(page_lru, &zone->inactive_dirty_list);
				UnlockPage(page);
				continue;
			}
		}

		/*
		 * Dirty swap-cache page? Write it out if
		 * last copy..
		 */
		if (PageDirty(page)) {
			int (*writepage)(struct page *);

			/* Can a page get here without page->mapping? */
			if (!page->mapping)
				goto page_active;
			writepage = page->mapping->a_ops->writepage;
			if (!writepage)
				goto page_active;
				
			/* Move it to the back of the list since either we want to skip it
			 * (first time through) or we writepage() it and then we don't want
			 * to see it the next while iteration 
			 */
		
			list_del(page_lru);
			list_add(page_lru, &zone->inactive_dirty_list);


			/* First time through? */
			if (!launder_loop || !CAN_DO_FS) {
				UnlockPage(page);
				continue;
			}

			if (vmfixes && atomic_read(&queued_sectors) > high_queued_sectors) {
				UnlockPage(page);
				continue;
			}

			/* OK, do a physical asynchronous write to swap.  */
			ClearPageDirty(page);
			page_cache_get(page);
			spin_unlock(&pagemap_lru_lock);

			writepage(page);
			page_cache_release(page);

			/* And re-start the thing.. */
			spin_lock(&pagemap_lru_lock);
			continue;
		}

		/*
		 * If the page has buffers, try to free the buffer mappings
		 * associated with this page. If we succeed we either free
		 * the page (in case it was a buffercache only page) or we
		 * move the page to the inactive_clean list.
		 *
		 * On the first round, we should free all previously cleaned
		 * buffer pages
		 */
		if (page->buffers) {
			unsigned int buffer_mask;
			int clearedbuf;
			int freed_page = 0;
			/*
			 * Since we might be doing disk IO, we have to
			 * drop the spinlock and take an extra reference
			 * on the page so it doesn't go away from under us.
			 */
			del_page_from_inactive_dirty_list(page);
			page_cache_get(page);
			spin_unlock(&pagemap_lru_lock);

			/* Will we do (asynchronous) IO? */
			if (launder_loop && maxlaunder == 0 && sync)
				buffer_mask = gfp_mask;				/* Do as much as we can */
			else if (launder_loop && maxlaunder-- > 0)
				buffer_mask = gfp_mask & ~__GFP_WAIT;			/* Don't wait, async write-out */
			else
				buffer_mask = gfp_mask & ~(__GFP_WAIT | __GFP_IO);	/* Don't even start IO */

			/* Try to free metadata underlying the page. */
			clearedbuf = try_to_release_page(page, buffer_mask);

			/*
			 * Re-take the spinlock. Note that we cannot
			 * unlock the page yet since we're still
			 * accessing the page_struct here...
			 */
			spin_lock(&pagemap_lru_lock);

			/* The buffers were not freed. */
			if (!clearedbuf) {
				add_page_to_inactive_dirty_list(page);

			/* The page was only in the buffer cache. */
			} else if (!page->mapping) {
				atomic_dec(&buffermem_pages);
				freed_page = 1;
				cleaned_pages++;

			/* The page has more users besides the cache and us. */
			} else if (page_count(page) > 2) {
				add_page_to_active_list(page);

			/* OK, we "created" a freeable page. */
			} else /* page->mapping && page_count(page) == 2 */ {
				add_page_to_inactive_clean_list(page);
				cleaned_pages++;
			}

			/*
			 * Unlock the page and drop the extra reference.
			 * We can only do it here because we are accessing
			 * the page struct above.
			 */
			UnlockPage(page);
			page_cache_release(page);

			/* 
			 * If we're freeing buffer cache pages, stop when
			 * we've got enough free memory.
			 */
			if (freed_page && !free_shortage())
				break;

			continue;
		} else if (page->mapping && !PageDirty(page)) {
			/*
			 * If a page had an extra reference in
			 * deactivate_page(), we will find it here.
			 * Now the page is really freeable, so we
			 * move it to the inactive_clean list.
			 */
			del_page_from_inactive_dirty_list(page);
			add_page_to_inactive_clean_list(page);
			UnlockPage(page);
			cleaned_pages++;
		} else {
page_active:
			/*
			 * OK, we don't know what to do with the page.
			 * It's no use keeping it here, so we move it to
			 * the active list.
			 */
			del_page_from_inactive_dirty_list(page);
			add_page_to_active_list(page);
			page->age = max((int)page->age, vm_age_parms.start);
			UnlockPage(page);
		}
	}
	spin_unlock(&pagemap_lru_lock);

	/*
	 * If we don't have an abundance of free pages, we loop back once
	 * to queue the dirty pages for writeout. When we were called
	 * by a user process (that /needs/ a free page) and we didn't
	 * free anything yet, we wait synchronously on the writeout of
	 * MAX_SYNC_LAUNDER pages.
	 *
	 * We also wake up bdflush, since bdflush should, under most
	 * loads, flush out the dirty pages before we have to wait on
	 * IO.
	 */
	if (CAN_DO_IO && !launder_loop && free_high(zone)>0) {
		launder_loop = 1;
		/* If we cleaned pages, never do synchronous IO. */
		if (cleaned_pages)
			sync = 0;
		/* We only do a few "out of order" flushes. */
		maxlaunder = MAX_LAUNDER;
		/* signal Kflushd to take care of the rest. */
		want_bdflush = 1;
		goto dirty_page_rescan;
	}
	if (want_bdflush)
		wakeup_bdflush(0);

	/* Return the number of pages moved to the inactive_clean list. */
	return cleaned_pages;
}

#ifdef CONFIG_HIGHMEM
extern int zone_page_count_type(int);
extern int zone_free_shortage_type(int);
extern int slabpages;
#define zone_is_highmem(z) (PageHighMem(mem_map + (z)->zone_start_mapnr))
#define low_mem_slab_excess() ((slabpages > (zone_page_count_type(ZONE_NORMAL)/2)) \
		&& zone_free_shortage_type(ZONE_NORMAL))
#endif

int page_launder(int gfp_mask, int sync)
{
        int maxtry = 1 << DEF_PRIORITY;
        struct zone_struct * zone;
        pg_data_t * pgdat;
        int freed = 0;

        /* Global balancing while we have a global shortage. */
        while (maxtry-- && free_high(ALL_ZONES) >= 0) {
                for_each_zone(pgdat, zone)
                        if (free_plenty(zone) >= 0)
                                freed += page_launder_zone(zone, gfp_mask, sync, 0);
        }

        /* Clean up the remaining zones with a serious shortage, if any. */
        for_each_zone(pgdat, zone)
                if ((free_min(zone) >= 0)
#ifdef CONFIG_HIGHMEM
			/* If there is lowmem pressure due to excessive buffer_headers
			 * in the slab which is caused by a large highmem pagecache then 
			 * launder the highmem zone so try_to_free_buffers runs */
			       	|| (zone_is_highmem(zone) && low_mem_slab_excess())
#endif
		   )
                        freed += page_launder_zone(zone, gfp_mask, sync, 1);

        return freed;
}

/* Check to see if a given page is a swap cache page with no further
 * references to the swap entry nor to the page in memory.  Clear all of
 * the aging information for such pages to force them immediately onto
 * the inactive_clean list for page reclamation. */

static void check_orphaned_swap(struct page *page)
{
	spinlock_t *pg_lock = PAGECACHE_LOCK(page);

	if (spin_trylock(pg_lock)) {
		/* Re-test the page state under protection of the proper
		 * spinlock */
		if (PageSwapCache(page) && atomic_read(&page->count) == 1+!!(page->buffers)) {
			if (!TryLockPage(page)) {
				if (swap_count(page) == 1) {
					ClearPageDirty(page);
					ClearPageReferenced(page);
					page->age = 0;
				}
				UnlockPage(page);
			}
		}
		spin_unlock(pg_lock);
	}
}



/**
 * refill_inactive_scan - scan the active list and find pages to deactivate
 * @priority: the priority at which to scan
 * @target: number of pages to deactivate, zero for background aging
 *
 * This function will scan a portion of the active list to find
 * unused pages, those pages will then be moved to the inactive list.
 */
static int refill_inactive_zone(struct zone_struct * zone, unsigned int priority)
{
	struct list_head * page_lru;
	struct page * page;
	int maxscan = zone->active_pages >> priority;
	int nr_deactivated = 0;

	/* Take the lock while messing with the list... */
	spin_lock(&pagemap_lru_lock);
	while (maxscan-- > 0 && (page_lru = zone->active_list.prev) != &zone->active_list) {
		page = list_entry(page_lru, struct page, lru);

		/* Wrong page on list?! (list corruption, should not happen) */
		if (!PageActive(page)) {
			printk("VM: refill_inactive, wrong page on list.\n");
			list_del(page_lru);
			nr_active_pages--;
			continue;
		}
	
		/* Special case: orphaned swap cache pages should be
		 * reclaimed as quickly as possible, regardless of their
		 * age or whether they are dirty or not. */
		if (PageSwapCache(page) && atomic_read(&page->count) == 1+!!(page->buffers)) {
			check_orphaned_swap(page);
			if (!page->age)
				deactivate_page_nolock_lifo(page);
			/* also don't count this as work */
			maxscan++;
		}

		/* Do aging on the pages. */
		if (PageTestandClearReferenced(page)) {
			age_page_up(page);
		} else {
			age_page_down(page);
			if (!page->age)
				deactivate_page_nolock(page);
		}

		/*
		 * If the amount of buffer cache pages is too
		 * high we just move every buffer cache page we
		 * find to the inactive list. Eventually they'll
		 * be reclaimed there...
		 */
		if (page->buffers && !page->mapping && too_many_buffers())
			deactivate_page_nolock(page);

		/*
		 * If the page cache is too large, we deactivate all
		 * page cache pages which are not in use by a process.
		 */
		if (pagecache_too_large() && page->mapping &&
				page_count(page) <= (page->buffers ? 2 : 1))
			deactivate_page_nolock(page);

		/*
		 * If the page is still on the active list, move it
		 * to the other end of the list. Otherwise we exit if
		 * we have done enough work.
		 */
skip_page:
		if (PageActive(page)) {
			list_del(page_lru);
			list_add(page_lru, &zone->active_list);
		} else {
			nr_deactivated++;
		}
	}
	spin_unlock(&pagemap_lru_lock);

	return nr_deactivated;
}

int refill_inactive_scan(unsigned int priority)
{
	int maxtry = 1 << priority;
	pg_data_t * pgdat;
	zone_t * zone;
	int ret = 0;

	/* Global balancing while we have a global shortage. */
	while (maxtry-- && inactive_low(ALL_ZONES) >= 0) {
		for_each_zone(pgdat, zone) {
			if (inactive_high(zone) >= 0)
				ret += refill_inactive_zone(zone, priority);
		}

	}

	/* Local balancing for zones which really need it. */
	for_each_zone(pgdat, zone) {
		if (inactive_low(zone) >= 0)
			ret += refill_inactive_zone(zone, priority);
	}

	return ret;
}

void refill_inactive_backgroundscan(void)
{
	pg_data_t * pgdat;
	zone_t * zone;

	/* background aging of all zones */
	for_each_zone(pgdat, zone) {
		if (!zone_inactive_plenty(zone))
			refill_inactive_zone(zone, DEF_PRIORITY+3);
	}

}

static int scanpercent = 100;
static long lasttime = 0;

void page_steal_zone(zone_t * zone, int gfp_mask)
{
        int maxscan, scandepth;
        int progress = 0;
        struct list_head * page_lru;
        struct page * page;
        struct address_space * mapping;
        spinlock_t *pg_lock;

	spin_lock(&pagemap_lru_lock);
	scanpercent += min(((jiffies-lasttime)*100/HZ/10), 10);
	if (scanpercent > 100)
		scanpercent = 100;
	lasttime = jiffies;
        maxscan = zone->inactive_dirty_pages*scanpercent/100;
	spin_unlock(&pagemap_lru_lock);
	scandepth = maxscan;
retry1:
	spin_lock(&pagemap_lru_lock);
        while ((page_lru = zone->inactive_dirty_list.prev) != &zone->inactive_dirty_list
                && pagecache_over_max() && --scandepth>0) {

                page = list_entry(page_lru, struct page, lru);

                pg_lock = PAGECACHE_LOCK(page);
                if (!spin_trylock(pg_lock)) {
                        spin_unlock(&pagemap_lru_lock);
                        goto retry1;
                }

		if (PageSwapCache(page) || TryLockPage(page)) {
			list_del(page_lru);
			list_add(page_lru, &zone->inactive_dirty_list);
			spin_unlock(pg_lock);
			continue;
		}

                if (!page->mapping || page_count(page) > (page->buffers ? 2 : 1)) {
                        list_del(page_lru);
                        list_add(page_lru, &zone->inactive_dirty_list);
			UnlockPage(page);
                        spin_unlock(pg_lock);
                        continue;
                }

                if (PageDirty(page)) {
                        list_del(page_lru);
                        list_add(page_lru, &zone->inactive_dirty_list);
			UnlockPage(page);
                        spin_unlock(pg_lock);
                        continue;
                }

                page_cache_get(page);

                if (page->buffers) { 
		    	if (!try_to_release_page(page, gfp_mask & ~(__GFP_WAIT | __GFP_IO))) {
                                list_del(page_lru);
                                list_add(page_lru, &zone->inactive_dirty_list);
				UnlockPage(page);
                        	spin_unlock(pg_lock);
                                page_cache_release(page);
				wakeup_bdflush(0);
                                continue;
                        }
                } 
		mapping = page->mapping;
		spin_lock(&mapping->page_lock);
		__remove_inode_page(page);
		spin_unlock(&mapping->page_lock);
		page_cache_release(page);

		UnlockPage(page);
                spin_unlock(pg_lock);
                del_page_from_inactive_dirty_list(page);
                page_cache_release(page);
                progress++;
        }
	if (scandepth >= maxscan*9/10) {
		scanpercent += (scandepth - maxscan*9/10)*100/max(1,maxscan);
		if (scanpercent > 100)
			scanpercent = 100;
	} else {
		scanpercent -= 10;
		if (scanpercent < 10)
			scanpercent = 10;
	}
        spin_unlock(&pagemap_lru_lock);

retry2:
        spin_lock(&pagemap_lru_lock);
        maxscan = zone->inactive_clean_pages;
        while ((page_lru = zone->inactive_clean_list.prev) != &zone->inactive_clean_list
                && pagecache_over_max() && maxscan-- > 0) {

                page = list_entry(page_lru, struct page, lru);

                pg_lock = PAGECACHE_LOCK(page);
                if (!spin_trylock(pg_lock)) {
                        spin_unlock(&pagemap_lru_lock);
                        goto retry2;
                }

		if (PageSwapCache(page) || TryLockPage(page)) {
			list_del(page_lru);
			list_add(page_lru, &zone->inactive_clean_list);
			spin_unlock(pg_lock);
			continue;
		}

                if (!page->mapping || page_count(page) > (page->buffers ? 2 : 1)) {
                        list_del(page_lru);
                        list_add(page_lru, &zone->inactive_clean_list);
			UnlockPage(page);	
                        spin_unlock(pg_lock);
                        continue;
                }

                if (PageDirty(page)) {
                        del_page_from_inactive_clean_list(page);
                        add_page_to_inactive_dirty_list(page);
			UnlockPage(page);
                        spin_unlock(pg_lock);
                        continue;
                }

                page_cache_get(page);

                if (page->buffers) {
                        if (!try_to_release_page(page, gfp_mask & ~(__GFP_WAIT | __GFP_IO))) {
				del_page_from_inactive_clean_list(page);
				add_page_to_inactive_dirty_list(page);
				UnlockPage(page);
                        	spin_unlock(pg_lock);
                                page_cache_release(page);
				wakeup_bdflush(0);
                                continue;
                        }
                }
		mapping = page->mapping;
		spin_lock(&mapping->page_lock);
		__remove_inode_page(page);
		spin_unlock(&mapping->page_lock);
		page_cache_release(page);
	
		UnlockPage(page);		
                spin_unlock(pg_lock);
                del_page_from_inactive_clean_list(page);
                page_cache_release(page);
                progress++;
        }
        spin_unlock(&pagemap_lru_lock);

        if (progress)
                refill_inactive_zone(zone, DEF_PRIORITY+3);
}

long count_ramdisk_pages(void)
{
	struct list_head *page_lru;
	struct page *page;
	long nr_ramdisk = 0;
        pg_data_t * pgdat;
	zone_t * zone;

	if (!spin_trylock(&pagemap_lru_lock)) {
		printk("count_ramdisk_pages: pagemap_lru_lock locked\n");
		return 0;
	}
	for_each_zone(pgdat, zone) {
		for (page_lru = zone->active_list.next; page_lru != &zone->active_list;
		     page_lru = page_lru->next) {
			page = list_entry(page_lru, struct page, lru);
			if (page_ramdisk(page))
				nr_ramdisk ++;
		}
	}
	spin_unlock(&pagemap_lru_lock);

	return nr_ramdisk;
}

/*
 * Check if there are zones with a severe shortage of free pages,
 * or if all zones have a minor shortage.
 */
int free_shortage(void)
{
	pg_data_t *pgdat;
	unsigned int global_free = 0;
	unsigned int global_target = freepages.high;

	/* Are we low on free pages anywhere? */
	pgdat = pgdat_list;
	do {
		int i;
		for(i = 0; i < MAX_NR_ZONES; i++) {
			zone_t *zone = pgdat->node_zones+ i;
			unsigned int free;

			if (!zone->size)
				continue;

			free = zone->free_pages;
			free += zone->inactive_clean_pages;

			/* Local shortage? */
			if (free < zone->pages_low)
				return 1;

			global_free += free;
		}
		pgdat = pgdat->node_next;
	} while (pgdat);

	/* Global shortage? */
	return global_free < global_target;
}

/*
 * Are we low on inactive pages globally or in any zone?
 */
int inactive_shortage(void)
{
	pg_data_t *pgdat;
	unsigned int global_target = freepages.high + inactive_target();
	unsigned int global_inactive = 0;

	pgdat = pgdat_list;
	do {
		int i;
		for(i = 0; i < MAX_NR_ZONES; i++) {
			zone_t *zone = pgdat->node_zones + i;
			unsigned int inactive;

			if (!zone->size)
				continue;

			inactive  = zone->inactive_dirty_pages;
			inactive += zone->inactive_clean_pages;
			inactive += zone->free_pages;

			/* Local shortage? */
			if (inactive < zone->pages_high)
				return 1;

			global_inactive += inactive;
		}
		pgdat = pgdat->node_next;
	} while (pgdat);

	/* Global shortage? */
	return global_inactive < global_target;
}

#define DEF_PRIORITY (6)
#define LOW_PRIORITY (10)

static inline int swapout_priority(void)
{
	int priority;
	int non_anon, non_anon_ratio;
	extern int boot_bigpages_reserved;

	if (pagecache_too_large() || too_many_buffers())
		return LOW_PRIORITY;
	else {
		non_anon = atomic_read(&buffermem_pages) + atomic_read(&page_cache_size) 
			   - swapper_space.nrpages + boot_bigpages_reserved
			   + total_reserved_pages;
		non_anon_ratio = non_anon*100/num_physpages;
		priority = DEF_PRIORITY+((LOW_PRIORITY-DEF_PRIORITY)*non_anon_ratio/100);
		return priority;
	}
}
/*
 * Refill_inactive is the function used to scan and age the pages on
 * the active list and in the working set of processes, moving the
 * little-used pages to the inactive list.
 *
 * When called by kswapd, we try to deactivate as many pages as needed
 * to recover from the inactive page shortage. This makes it possible
 * for kswapd to keep up with memory demand so user processes can get
 * low latency on memory allocations.
 *
 * However, when the system starts to get overloaded we can get called
 * by user processes. For user processes we want to both reduce the
 * latency and make sure that multiple user processes together don't
 * deactivate too many pages. To achieve this we simply do less work
 * when called from a user process.
 */
static int refill_inactive(unsigned int gfp_mask)
{
	int progress = 0, maxtry;

	maxtry = 1 << DEF_PRIORITY;

	do {
		if (current->need_resched) {
			 __set_current_state(TASK_RUNNING);
			schedule();
			if (!inactive_shortage())
				return 1;
		}

		/* Walk the VM space for a bit.. */
		swap_out((vmfixes?swapout_priority():DEF_PRIORITY), gfp_mask);

		/* ..and refill the inactive list */
		progress += refill_inactive_scan(DEF_PRIORITY);

		if (--maxtry <= 0)
			break;
	} while (inactive_shortage());

	return progress;
}

#ifdef CONFIG_SMP 
DECLARE_WAIT_QUEUE_HEAD(krefilld_wait);
#endif

/*
 * Worker function for kswapd and try_to_free_pages, we get
 * called whenever there is a shortage of free/inactive_clean
 * pages.
 *
 * This function will also move pages to the inactive list,
 * if needed.
 */
static int do_try_to_free_pages(unsigned int gfp_mask, int user)
{
	int ret = 0;

	/*
	 * Eat memory from filesystem page cache, buffer cache,
	 * dentry, inode and filesystem quota caches.
	 */
	ret += page_launder(gfp_mask, user);
	ret += shrink_dcache_memory(3, gfp_mask);
	/* look at 1% of the inode cache */
	ret += shrink_icache_memory(100, gfp_mask);
	ret += shrink_dqcache_memory(DEF_PRIORITY, gfp_mask);

	/*
	 * If needed, we move pages from the active list
	 * to the inactive list.
	 */
	if (inactive_shortage()) {
#ifdef CONFIG_SMP
		if (vmfixes && smp_num_cpus > 1 && !user)
			wake_up_interruptible(&krefilld_wait);
		else
			ret += refill_inactive(gfp_mask);
#else
		ret += refill_inactive(gfp_mask);
#endif
	}

	/* 	
	 * Reclaim unused slab cache memory.
	 */
	kmem_cache_reap(gfp_mask);

	return ret;
}

DECLARE_WAIT_QUEUE_HEAD(kswapd_wait);
DECLARE_WAIT_QUEUE_HEAD(kswapd_done);
static void wakeup_memwaiters(void);

/*
 * The background pageout daemon, started as a kernel thread
 * from the init process. 
 *
 * This basically trickles out pages so that we have _some_
 * free memory available even if there is no other activity
 * that frees anything up. This is needed for things like routing
 * etc, where we otherwise might have all activity going on in
 * asynchronous contexts that cannot page things out.
 *
 * If there are applications that are active memory-allocators
 * (most normal use), this basically shouldn't matter.
 */
int kswapd(void *unused)
{
	struct task_struct *tsk = current;

	daemonize();
	strcpy(tsk->comm, "kswapd");
	sigfillset(&tsk->blocked);
	
	/*
	 * Tell the memory management that we're a "memory allocator",
	 * and that if we need more memory we should get access to it
	 * regardless (see "__alloc_pages()"). "kswapd" should
	 * never get caught in the normal page freeing logic.
	 *
	 * (Kswapd normally doesn't need memory anyway, but sometimes
	 * you need a small amount of memory in order to be able to
	 * page out something else, and this flag essentially protects
	 * us from recursively trying to free more memory as we're
	 * trying to free the first piece of memory in the first place).
	 */
	tsk->flags |= PF_MEMALLOC;

	/*
	 * Kswapd main loop.
	 */
	for (;;) {
		static long recalc = 0;
		static long recalc2 = 0;

		/* If needed, try to free some memory. */
		if (free_high(ALL_ZONES) >= 0 || free_low(ANY_ZONE) > 0)
			do_try_to_free_pages(GFP_KSWAPD, 0);

		/* Once every 15 seconds, 5% results in everything in 5 minutes... */
		if (time_after(jiffies, recalc2 + 15* HZ)) {
			recalc2 = jiffies;
			shrink_dcache_memory(20,GFP_KSWAPD);
			shrink_icache_memory(20,GFP_KSWAPD);
		}
	
		/* Once a second ... */
		if (time_after(jiffies, recalc + HZ)) {
			recalc = jiffies;

			/* Do background page aging. */
			refill_inactive_backgroundscan();
		}

		wakeup_memwaiters();
	
		/* 
		 * We go to sleep if either the free page shortage
		 * or the inactive page shortage is gone. We do this
		 * because:
		 * 1) we need no more free pages   or
		 * 2) the inactive pages need to be flushed to disk,
		 *    it wouldn't help to eat CPU time now ...
		 *
		 * We go to sleep for one second, but if it's needed
		 * we'll be woken up earlier...
		 */
		if (!free_shortage() || !inactive_shortage()) {
			interruptible_sleep_on_timeout(&kswapd_wait, HZ);
		/*
		 * If we couldn't free enough memory, we see if it was
		 * due to the system just not having enough memory.
		 * If that is the case, the only solution is to kill
		 * a process (the alternative is enternal deadlock).
		 *
		 * If there still is enough memory around, we just loop
		 * and try free some more memory...
		 */
		} else if (out_of_memory()) {
			oom_kill();
		}
	}
}

spinlock_t vm_sleep_lock ____cacheline_aligned_in_smp = SPIN_LOCK_UNLOCKED;
int vm_sleep_time = 0;
#define VMMAXSLEEP 30

#define VM_SHOULD_SLEEP (free_low(ALL_ZONES) > (int)(freepages.min / 2))
static int kswapd_overloaded = 0;

/**
 * wakeup_kswapd - wake up the pageout daemon
 * gfp_mask: page freeing flags
 *
 * This function wakes up kswapd and can, under heavy VM pressure,
 * put the calling task to sleep temporarily.
 */
void wakeup_kswapd(unsigned int gfp_mask)
{
	DECLARE_WAITQUEUE(wait, current);


	/* If we're in the memory freeing business ourself, don't sleep */
	if (current->flags & PF_MEMALLOC) {
		wake_up_interruptible(&kswapd_wait);
		return;
	}

	/* We need all of kswapd's GFP flags, otherwise we can't sleep on it. */
	if ((gfp_mask & GFP_KSWAPD) != GFP_KSWAPD) {	
		wake_up_interruptible(&kswapd_wait);
		return;
	}
	
	add_wait_queue(&kswapd_done, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);

	wake_up_interruptible(&kswapd_wait);

	/* Sleep until we're woken up by kswapd below. */
	if (VM_SHOULD_SLEEP && !(kswapd_overloaded)) {
		spin_lock(&vm_sleep_lock);
		vm_sleep_time = VMMAXSLEEP;
		spin_unlock(&vm_sleep_lock);
		schedule();
	}
nosleep:
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&kswapd_done, &wait);

}

static void wakeup_memwaiters(void)
{
	int vm_slept_too_long = 0;

	spin_lock(&vm_sleep_lock);
	if (vm_sleep_time) 
		vm_slept_too_long = (--vm_sleep_time==0?1:0);

	/* Enough free RAM, we can easily keep up with memory demand. */
	if (free_high(ALL_ZONES) <= 0) {
		vm_sleep_time = 0;
		spin_unlock(&vm_sleep_lock);
		wake_up(&kswapd_done);
		interruptible_sleep_on_timeout(&kswapd_wait, HZ);
		return;
	}

	/* If the sleeping condition is over ... */
	if (!VM_SHOULD_SLEEP || vm_slept_too_long) {
		vm_sleep_time = 0;
		spin_unlock(&vm_sleep_lock);
		wake_up(&kswapd_done);
	} else
		spin_unlock(&vm_sleep_lock);

	/* OK, the VM is very loaded. Sleep instead of using all CPU. */
	kswapd_overloaded = 1;
	set_current_state(TASK_UNINTERRUPTIBLE);
	schedule_timeout(HZ / 40);
	kswapd_overloaded = 0;
	wmb();
                                               	
	return;
}

/*
 * Called by non-kswapd processes when they want more
 * memory but are unable to sleep on kswapd because
 * they might be holding some IO locks ...
 */
int try_to_free_pages(unsigned int gfp_mask)
{
	int ret = 1;

	if (gfp_mask & __GFP_WAIT) {
		unsigned long flags;
		flags =current->flags;
		current->flags |= PF_MEMALLOC;
		ret = do_try_to_free_pages(gfp_mask, 1);
		current->flags = (current->flags & ~PF_MEMALLOC) | (flags & PF_MEMALLOC);
	}

	return ret;
}

DECLARE_WAIT_QUEUE_HEAD(kreclaimd_wait);
/*
 * Kreclaimd will move pages from the inactive_clean list to the
 * free list, in order to keep atomic allocations possible under
 * all circumstances.
 */
int kreclaimd(void *unused)
{
	struct task_struct *tsk = current;
	pg_data_t *pgdat;

	daemonize();
	strcpy(tsk->comm, "kreclaimd");
	sigfillset(&tsk->blocked);
	current->flags |= PF_MEMALLOC;

	while (1) {

		/*
		 * We sleep until someone wakes us up from
		 * page_alloc.c::__alloc_pages().
		 */
		interruptible_sleep_on(&kreclaimd_wait);

		/*
		 * Move some pages from the inactive_clean lists to
		 * the free lists, if it is needed.
		 */
		pgdat = pgdat_list;
		do {
			int i;
			for(i = 0; i < MAX_NR_ZONES; i++) {
				zone_t *zone = pgdat->node_zones + i;
				if (!zone->size)
					continue;

				while (zone->free_pages < zone->pages_low) {
					struct page * page;
					page = reclaim_page(zone);
					if (!page)
						break;
					__free_page(page);
				}
			}
			pgdat = pgdat->node_next;
		} while (pgdat);
	}
}

#ifdef CONFIG_SMP 
int krefilld(void *unused)
{
        struct task_struct *tsk = current;

        daemonize();
        strcpy(tsk->comm, "krefilld");
        sigfillset(&tsk->blocked);
        current->flags |= PF_MEMALLOC;

        while (1) {

                /*
                 * We sleep until kswaod wakes us up from
		 * do_try_to_free_pages
                 */
                interruptible_sleep_on(&krefilld_wait);

		refill_inactive(GFP_KSWAPD);

        }
}
#endif

static int __init kswapd_init(void)
{
	printk("Starting kswapd v1.8\n");
	swap_setup();
	kernel_thread(kswapd, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGNAL);
	kernel_thread(kreclaimd, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGNAL);
#ifdef CONFIG_SMP
	if (vmfixes)
		kernel_thread(krefilld, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGNAL);
#endif
	return 0;
}

static int __init vmfixes_setup(char *str)
{
	vmfixes = 0;
	return 1;
}

__setup("novmfixes", vmfixes_setup);

module_init(kswapd_init)
